from .nn.op import HMMlayer
import pickle
import re, sys, os

class SegModel():
    def __init__(self, model_path):
        model = open(model_path, 'rb')
        self.transitionProb = pickle.load(model)
        self.emissionProb = pickle.load(model)
        self.word_list = pickle.load(model)
        self.pi = pickle.load(model)
        self.states = pickle.load(model)
        self.word_dict = pickle.load(model)

    def catchStr(self, sentence):
        l=[]
        l=re.split('(,|。|、|———|/|●)', sentence)
        values = l[::2]
        delimiters = l[1::2] + ['']

        return [v + d for v, d in zip(values, delimiters)]

    def convertSentence(self, sentence):
        l=[]
        for word in  sentence:
            try:
                l.append(self.word_list.index(word))
            except ValueError:
                l.append(self.word_list.index('。'))

        return l

    def outPutResult(self, sentence, s_seq, term_list, states):
        s = ''
        for i in range(len(sentence)):
            tag = states[s_seq[i]]
            if tag == 'E' or tag == 'S':
                # print(term_list[o_seq[i]], end=' ')
                s += sentence[i] + " "
            else:
                # print(term_list[o_seq[i]], end='')
                s += sentence[i]
        return s

    def mfm(self, string, max_len=5):
        """
        Max forward match
        """

        def getSeg(text):
            if not text:
                return ''
            if len(text) == 1:
                return text
            if text in self.word_dict:
                return text
            else:
                small = len(text) - 1
                text = text[0:small]
                return getSeg(text)

        max_len = max_len
        result_str = ''
        result_len = 0
        unk = ''
        while string:
            tmp_str = string[0:max_len]
            seg_str = getSeg(tmp_str)
            seg_len = len(seg_str)
            if seg_len == 1:
                unk += seg_str
            else:
                if unk != '':
                    s = self.hmm(unk)
                    result_str = result_str + s
                    unk = ''
            result_len = result_len + seg_len

            if seg_str.strip() and seg_len != 1:
                result_str = result_str + seg_str + ' '
            string = string[seg_len:]
        result_str = result_str + ' ' + self.hmm(unk)

        return result_str




    def hmm(self, sentence):  # sentence 为分词后的数组形式

        s=''
        # print(self.word_list)
        windows = self.catchStr(sentence)
        # print(windows)
        for s_window in windows:

            if s_window=='':
                continue
            o_seq = self.convertSentence(s_window+"。")
            s_seq=HMMlayer(o_seq, self.transitionProb, self.emissionProb, self.pi)
            s += self.outPutResult(s_window, s_seq, self.word_list, self.states)
        return s

    def cut(self, string):
        result = self.mfm(string)
        return result


import numpy as np


class Tensor:
    value = None
    shape = None

    def __init__(self, data=None):
        self.value = np.array(data)

    def __iter__(self):
        self.iter = 0
        return self

    def __next__(self):

        if self.iter+1 < self.shape()[0]:
            self.iter += 1
        return self.value[self.iter]

    def __add__(self, tensor):
        return Tensor(np.add(self.value, tensor.value))

    def __mul__(self, tensor):
        return Tensor(np.multiply(self.value, tensor.value))

    def reshape(self, shape):
        self.value = np.reshape(self.value, shape)
        return self

    def shape(self):
        return self.value.shape

    @staticmethod
    def zero(shape):
        return Tensor(np.zeros(shape))

    @staticmethod
    def ones(shape):
        return Tensor(np.ones(shape))



if __name__ == "__main__":
    a = Tensor([[1, 2, 3], [4, 5, 6]])
    b = Tensor([[7, 8, 9], [3, 4, 5]])
    c = Tensor.ones([2, 2])
    d = a + b
    e = a.reshape([3, 2])
    for i in a:
        print(i)


from .util import parser_parameters
import numpy as np


class ParserModel:
    def parser(self):
        def __init__(self):
            self.paras = parser_parameters("path")
            self.embedding = self.paras['embeding']
            self.word2id = self.paras["word2id"]
            self.id2tag = self.paras["id2tag"]
            self.Wi = np.array(self.paras["Wi"])
            self.bi = np.array(self.paras["bi"])
            self.Wf = np.array(self.paras["Wf"])
            self.bf = np.array(self.paras["bf"])
            self.Wu = np.array(self.paras["Wu"])
            self.bu = np.array(self.paras["bu"])
            self.Wo = np.array(self.paras["Wo"])
            self.bo = np.array(self.paras["bo"])
            self.W_tag = np.array(self.paras["W_tag"])
            self.trans = np.array(self.paras["transitions"])

        def parser():
            pass



import numpy as np


def lookup(word_list, vocab, word2id):
    # ids = [word2id[word] if word in word2id else word2id['UNK'] for word in word_list]
    ids = [word2id[word] for word in word_list]
    emb = []
    for i in ids:
        emb.append(vocab[i])
    return emb

class LSTMparam:
    def __init__(self, wg, wi, wf, wo, bg, bi, bf, bo):
        # weight matrices
        self.wg = wg
        self.wi = wi
        self.wf = wf
        self.wo = wo
        # bias terms
        self.bg = bg
        self.bi = bi
        self.bf = bf
        self.bo = bo


class LSTMstate:
    def __init__(self, hidden_size):
        self.g = np.zeros(hidden_size)
        self.i = np.zeros(hidden_size)
        self.f = np.zeros(hidden_size)
        self.o = np.zeros(hidden_size)
        self.s = np.zeros(hidden_size)
        self.h = np.zeros(hidden_size)



def LSTMlayer(emb, Wi, bi, Wf, bf, Wg, bg, Wo, bo):

    h_pre = np.zeros((100))
    cell_state = np.zeros(bi.shape)
    out = []
    for xi in emb:
        xi = np.array(xi)
        input = sigmoid(np.dot(Wi, concat(xi, h_pre)) + bi)
        forget = sigmoid(np.dot(Wf, concat(xi, h_pre)) + bf)
        cell_update = tanh(np.dot(Wg, concat(xi, h_pre)) + bg)
        cell_state = (forget * cell_state) + (input * cell_update)
        output = sigmoid(np.dot(Wo, concat(xi, h_pre)) + bo)
        h_out = (tanh(output) * cell_state)
        out.append(h_out)
        h_pre = h_out
    return np.array(out)


class LSTMcell:
    def __init__(self, lstm_param, lstm_state):
        # store reference to parameters and to activations
        self.state = lstm_state
        self.param = lstm_param
        # non-recurrent input concatenated with recurrent input
        self.xc = None

    def update_cell(self, x, s_prev=None, h_prev=None):
        # if this is the first lstm node in the network
        if s_prev is None: s_prev = np.zeros_like(self.state.s)
        if h_prev is None: h_prev = np.zeros_like(self.state.h)
        # save data for use in backprop
        self.s_prev = s_prev
        self.h_prev = h_prev

        # concatenate x(t) and h(t-1)
        xc = np.hstack((x, h_prev))
        self.state.g = np.tanh(np.dot(self.param.wg) + self.param.bg)
        self.state.i = sigmoid(np.dot(self.param.wi, xc) + self.param.bi)
        self.state.f = sigmoid(np.dot(self.param.wf, xc) + self.param.bf)
        self.state.o = sigmoid(np.dot(self.param.wo, xc) + self.param.bo)
        self.state.s = self.state.g * self.state.i + s_prev * self.state.f
        self.state.h = self.state.s * self.state.o

        self.xc = xc

def Linear(input, W, b):
    return np.dot(input, W) + b

def CRFlayer(B, A):
    """
    Using viterbi algorithm to decoder
    :param B: LSTM prediction output
    :param A: The transition probability matrix of tags
    :return: the final score through CRF layer of each tag of each word
    """
    delta = np.zeros(B.shape)
    psi = np.zeros(B.shape)     # 记录路径
    for i in range(B.shape[1]):
        delta[0][i] = B[0][i]

    for t in range(1, B.shape[0]):
        for i in range(B.shape[1]):
            max = 0.0
            index = 0
            for j in range(B.shape[1]):
                if max < delta[t-1][j] + A[j][i] + B[t][i]:
                    max = delta[t-1][j] + A[j][i] + B[t][i]
                    index = j
            delta[t][i] = max
            psi[t][i] = index

    #   go back to find the maximum rotate
    max = 0
    max_id = 0
    t = len(delta) - 1
    for i in range(B.shape[1]):
        if delta[-1][i] > max:
            max = delta[-1][i]
            max_id = i
    state = [max_id]
    while t > 0:
        max_id = int(psi[t][max_id])
        state.append(max_id)
        t -= 1
    state.reverse()
    return state

def HMMlayer(o_sequence, A, B, pi):
    '''
        维特比算法：
        input:
            o_sequence:观测序列
            :条件转移概率
            :观测概率
            pi:初始状态概率

        return 符合最优的状态序列
    '''
    len_status = len(pi)

    status_record = {i: [[0, 0] for j in range(len(o_sequence))] for i in range(len_status)}

    for i in range(len(pi)):
        status_record[i][0][0] = pi[i] * B[i, o_sequence[0]]  # t时刻下状态为i的所有单个路径的最大概率,δt(i)=πi * bi(o1)
        status_record[i][0][1] = 0  # t时刻下状态为i的最大概率路径的第t-1个节点,Ψt(i)=0

    for t in range(1, len(o_sequence)):
        for i in range(len_status):
            max = [-1, 0]
            for j in range(len_status):  # max(δt(j)*aji) i = 1,2,...,N
                tmp_prob = status_record[j][t - 1][0] * A[j, i]  # δt(j) * A[j][i]
                if tmp_prob > max[0]:
                    max[0] = tmp_prob
                    max[1] = j

            status_record[i][t][0] = max[0] * B[i, o_sequence[t]]
            status_record[i][t][1] = max[1]

    max = 0
    max_idx = 0
    t = len(o_sequence) - 1
    for i in range(len_status):
        if max < status_record[i][t][0]:
            max = status_record[i][t][0]
            max_idx = i

    state_sequence = []  # 栈结构
    state_sequence.append(max_idx)
    while (t > 0):
        max_idx = status_record[max_idx][t][1]
        state_sequence.append(max_idx)
        t -= 1
    state_sequence.reverse()
    return state_sequence



def concat(tensor1, tensor2, axis=0):
    return np.concatenate((tensor1, tensor2), axis=axis)


def tanh(tensor):
    return np.tanh(tensor)


def sigmoid(tensor):
    for i in range(len(tensor)):
        tensor[i] = 1 / (1 + np.exp(-tensor[i]))
    return tensor



class Parameter():
    def __init__(self, dir):
        self.embedding = []
        self.word2id = []
        self.id2tag = []

    def load_vocab(self, dir):
        f = open(dir, 'r')
        for line in f:
            self.embedding.append(line)


#!/usr/bin/env python
# coding=utf-8

from setuptools import setup, find_packages

setup(
    name='Neu',
    version='1.4',
    description='中文自然语言处理工具包',
    long_description=open('README.rst').read(),
    autuor='wujinhang, huxiao',
    author_email='wujinhang0729@gmail.com',
    license='BSD License',
    # packages=find_packages(),
    url='https://github.com/Woooooody/Neu',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: Implementation',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries'
    ],
    keywords='NLP,tokenizing,Chinese word segementation',
    packages=['Neu'],
    package_dir={'Neu':'Neu'},
    package_data={'Neu':['*.*','model/*','nn/*']}

)



from .nn.op import lookup, LSTMlayer, CRFlayer, concat, Linear, LSTMcell, LSTMparam
from .util import ner_parameters
import numpy as np
import pickle
import os

class NerModel:
    def __init__(self, model_path):
        # self.paras = ner_parameters(os.getcwd()+"/Neu/model/ner.model")
        # paras = pickle.dump(self.paras, open('ner.model', 'wb'))
        # dir_path = os.path.dirname(os.path.abspath(__file__))
        # model_path = os.path.join(dir_path, 'model/ner.model')
        self.paras = pickle.load(open(model_path, 'rb'))
        self.embedding = self.paras['embeddings']
        self.word2id = self.paras["word2id"]
        self.id2tag = self.paras["id2tag"]
        #   forward lstm weight
        self.fWi = np.array(self.paras["fWi"]).reshape((100, 200))
        self.fWf = np.array(self.paras["fWf"]).reshape((100, 200))
        self.fWu = np.array(self.paras["fWu"]).reshape((100, 200))
        self.fWo = np.array(self.paras["fWo"]).reshape((100, 200))
        #   forward lstm bias
        self.fbi = np.array(self.paras["fbi"]).reshape((100))
        self.fbf = np.array(self.paras["fbf"]).reshape((100))
        self.fbu = np.array(self.paras["fbu"]).reshape((100))
        self.fbo = np.array(self.paras["fbo"]).reshape((100))
        #   backward lstm weight
        self.bWi = np.array(self.paras["bWi"]).reshape((100, 200))
        self.bWf = np.array(self.paras["bWf"]).reshape((100, 200))
        self.bWu = np.array(self.paras["bWu"]).reshape((100, 200))
        self.bWo = np.array(self.paras["bWo"]).reshape((100, 200))
        #   backward lstm bias
        self.bbi = np.array(self.paras["bbi"]).reshape((100))
        self.bbf = np.array(self.paras["bbf"]).reshape((100))
        self.bbu = np.array(self.paras["bbu"]).reshape((100))
        self.bbo = np.array(self.paras["bbo"]).reshape((100))
        #   linear weight and bias
        self.W_tag = np.array(self.paras["W_tag"]).reshape((200, 7))
        self.b_tag = np.array(self.paras["b_tag"]).reshape((7))
        #   transition matrix
        self.trans = np.array(self.paras["transitions"])

    def ner(self, string):
        string = string.replace(' ', '')
        words = [char for char in string]
        emb = lookup(words, self.embedding, self.word2id)   # (n, 100)
        forward = LSTMlayer(emb, self.fWi, self.fbi, self.fWf, self.fbf, self.fWu, self.fbu, self.fWo, self.fbo)
        emb.reverse()
        backward = LSTMlayer(emb, self.bWi, self.bbi, self.bWf, self.bbf, self.bWu, self.bbu, self.bWo, self.bbo)
        backward = np.flip(backward, 0)
        top_recur = concat(forward, backward, axis=1)
        tag_score = Linear(top_recur, self.W_tag, self.b_tag)
        tag_ids = CRFlayer(tag_score, self.trans)
        tag_list = [self.id2tag[id] for id in tag_ids]

        for i in range(len(words)):
            words[i] = words[i] + ' ' + tag_list[i]

        return words



import codecs
def ner_parameters(path):
    paras = {}
    keys = ['embeddings', 'fWi', 'fWu', 'fWf', 'fWo', 'fbi', 'fbu', 'fbf', 'fbo', 'bWi', 'bWu', 'bWf', 'bWo', 'bbi',
            'bbu', 'bbf', 'bbo', 'W_tag', 'b_tag', 'transitions', 'word2id', 'id2word']
    k = 0
    paras[keys[k]] = []
    file = codecs.open(path, 'r', encoding='utf-8')
    count = 0
    for line in file:
        if line.strip() == '#':
            k += 1
            paras[keys[k]] = []
            continue
        if line.strip() == '?':
            k += 1
            paras[keys[k]] = {}
            paras[keys[k + 1]] = {}
            continue
        tmp = line.split()
        if len(tmp) == 1:
            w2i = paras[keys[k]]
            i2w = paras[keys[k + 1]]
            w2i[tmp[0]] = count
            i2w[count] = tmp[0]
            count += 1
        else:
            tmp = [float(i) for i in tmp]
            paras[keys[k]].append(tmp)
    paras['id2tag'] = {0:'I-LOC', 1:'B-ORG', 2:'I-PER', 3:'O', 4:'I-ORG', 5:'B-LOC', 6:'B-PER'}
    file.close()
    return paras

def parser_parameters():
    pass

if __name__ == '__main__':
    paras = ner_parameters('ner.model')
    embeddings = paras['embeddings']
    word2id = paras['word2id']
    print(len(embeddings))



# -*- coding: utf-8 -*-
from .SegModel import SegModel
from .PosModel import PosModel
from .NerModel import NerModel
from .ParserModel import ParserModel
import os

dir_path = os.path.dirname(os.path.abspath(__file__))

cut = SegModel(dir_path+"/model/seg.model").cut
# pos = NerModel(dir_path+"/model/pos.model").ner
ner = NerModel(dir_path+"/model/ner.model").ner
# parser = ParserModel().parser


